package net.codinux.banking.client.fints4k

import net.codinux.banking.client.model.*
import net.codinux.banking.client.model.AccountTransaction
import net.codinux.banking.client.model.Amount
import net.codinux.banking.client.model.tan.*
import net.codinux.banking.client.model.options.GetAccountDataOptions
import net.codinux.banking.client.model.response.*
import net.codinux.banking.client.model.tan.ActionRequiringTan
import net.codinux.banking.client.model.tan.TanChallenge
import net.codinux.banking.client.model.tan.TanImage
import net.codinux.banking.client.model.tan.TanMethod
import net.codinux.banking.client.model.tan.TanMethodType
import net.dankito.banking.client.model.BankAccountIdentifierImpl
import net.dankito.banking.client.model.parameter.GetAccountDataParameter
import net.dankito.banking.client.model.parameter.RetrieveTransactions
import net.dankito.banking.client.model.response.ErrorCode
import net.dankito.banking.fints.mapper.FinTsModelMapper
import net.dankito.banking.fints.model.*
import kotlin.io.encoding.Base64
import kotlin.io.encoding.ExperimentalEncodingApi

open class FinTs4kMapper {

    private val fintsModelMapper = FinTsModelMapper()


    fun mapToGetAccountDataParameter(credentials: AccountCredentials, options: GetAccountDataOptions) = GetAccountDataParameter(
        credentials.bankCode, credentials.loginName, credentials.password,
        options.accounts.map { BankAccountIdentifierImpl(it.identifier, it.subAccountNumber, it.iban) },
        options.retrieveBalance,
        RetrieveTransactions.valueOf(options.retrieveTransactions.name), options.retrieveTransactionsFrom, options.retrieveTransactionsTo,
        abortIfTanIsRequired = options.abortIfTanIsRequired
    )


    fun map(response: net.dankito.banking.client.model.response.GetAccountDataResponse): Response<GetAccountDataResponse> {
        return if (response.successful && response.customerAccount != null) {
            Response.success(GetAccountDataResponse(mapCustomer(response.customerAccount!!)))
        } else {
            mapError(response)
        }
    }


    fun mapToCustomerAccountViewInfo(bank: BankData): CustomerAccountViewInfo = CustomerAccountViewInfo(
        bank.bankCode, bank.customerId, bank.bankName
    )

    fun mapToBankAccountViewInfo(account: AccountData): BankAccountViewInfo = BankAccountViewInfo(
        account.accountIdentifier, account.subAccountAttribute,
        mapAccountType(fintsModelMapper.map(account.accountType)),
        account.iban, account.productName
    )


    private fun mapCustomer(customer: net.dankito.banking.client.model.CustomerAccount): CustomerAccount = CustomerAccount(
        customer.bankCode, customer.loginName, customer.password,
        customer.bankName, customer.bic, customer.customerName, customer.userId,
        customer.accounts.map { mapAccount(it) }
    )


    private fun mapAccount(account: net.dankito.banking.client.model.BankAccount): BankAccount = BankAccount(
        account.identifier, account.accountHolderName, mapAccountType(account.type), account.iban, account.subAccountNumber,
        account.productName, account.currency, account.accountLimit, account.isAccountTypeSupportedByApplication,
        mapFeatures(account),
        mapAmount(account.balance), account.retrievedTransactionsFrom, account.retrievedTransactionsTo,
        // TODO: map haveAllTransactionsBeenRetrieved
        countDaysForWhichTransactionsAreKept = account.countDaysForWhichTransactionsAreKept,
        bookedTransactions = account.bookedTransactions.map { mapTransaction(it) }.toMutableList()
    )

    private fun mapAccountType(type: net.dankito.banking.client.model.BankAccountType): BankAccountType =
        BankAccountType.valueOf(type.name)

    private fun mapFeatures(account: net.dankito.banking.client.model.BankAccount): Set<BankAccountFeatures> = buildSet {
        if (account.supportsRetrievingBalance) {
            add(BankAccountFeatures.RetrieveBalance)
        }
        if (account.supportsRetrievingTransactions) {
            add(BankAccountFeatures.RetrieveTransactions)
        }
        if (account.supportsTransferringMoney) {
            add(BankAccountFeatures.TransferMoney)
        }
        if (account.supportsInstantPayment) {
            add(BankAccountFeatures.InstantPayment)
        }
    }


    private fun mapTransaction(transaction: net.dankito.banking.client.model.AccountTransaction): AccountTransaction = AccountTransaction(
        mapAmount(transaction.amount), transaction.amount.currency.code, transaction.unparsedReference,
        transaction.bookingDate, transaction.valueDate,
        transaction.otherPartyName, transaction.otherPartyBankCode, transaction.otherPartyAccountId,
        transaction.bookingText, null,
        transaction.statementNumber, transaction.sequenceNumber,
        mapNullableAmount(transaction.openingBalance), mapNullableAmount(transaction.closingBalance),
        // TODO: map other properties
    )

    private fun mapNullableAmount(amount: Money?) = amount?.let { mapAmount(it) }

    private fun mapAmount(amount: Money) = Amount.fromString(amount.amount.string.replace(',', '.'))


    fun mapTanChallenge(challenge: net.dankito.banking.fints.model.TanChallenge): TanChallenge {
        val type = mapTanChallengeType(challenge)
        val action = mapActionRequiringTan(challenge.forAction)
        val tanMethod = mapTanMethod(challenge.tanMethod)
        val customer = mapToCustomerAccountViewInfo(challenge.bank)
        val account = challenge.account?.let { mapToBankAccountViewInfo(it) }

        val tanImage = if (challenge is ImageTanChallenge) mapTanImage(challenge.image) else null
        val flickerCode = if (challenge is FlickerCodeTanChallenge) mapFlickerCode(challenge.flickerCode) else null

        return TanChallenge(type, action, challenge.messageToShowToUser, tanMethod, tanImage, flickerCode, customer, account)
    }

    private fun mapTanChallengeType(challenge: net.dankito.banking.fints.model.TanChallenge): TanChallengeType = when {
        challenge is ImageTanChallenge -> TanChallengeType.Image
        challenge is FlickerCodeTanChallenge -> TanChallengeType.Flickercode
        else -> TanChallengeType.EnterTan
    }

    private fun mapActionRequiringTan(action: net.dankito.banking.fints.model.ActionRequiringTan): ActionRequiringTan =
        ActionRequiringTan.valueOf(action.name)

    private fun mapTanMethod(method: net.dankito.banking.fints.model.TanMethod): TanMethod = TanMethod(
        method.displayName, mapTanMethodType(method.type), method.securityFunction.code, method.maxTanInputLength, mapAllowedTanFormat(method.allowedTanFormat)
    )

    private fun mapTanMethodType(type: net.dankito.banking.fints.model.TanMethodType): TanMethodType =
        TanMethodType.valueOf(type.name)

    private fun mapAllowedTanFormat(allowedTanFormat: net.dankito.banking.fints.messages.datenelemente.implementierte.tan.AllowedTanFormat?): AllowedTanFormat =
        allowedTanFormat?.let { AllowedTanFormat.valueOf(it.name) } ?: AllowedTanFormat.Alphanumeric

    private fun mapTanImage(image: net.dankito.banking.fints.tan.TanImage): TanImage =
        TanImage(image.mimeType, mapToBase64(image.imageBytes), mapException(image.decodingError))

    @OptIn(ExperimentalEncodingApi::class)
    private fun mapToBase64(bytes: ByteArray): String {
        return Base64.Default.encode(bytes)
    }

    private fun mapFlickerCode(flickerCode: net.dankito.banking.fints.tan.FlickerCode): FlickerCode =
        FlickerCode(flickerCode.challengeHHD_UC, flickerCode.parsedDataSet, mapException(flickerCode.decodingError))


    private fun <T> mapError(response: net.dankito.banking.client.model.response.GetAccountDataResponse): Response<T> {
        return if (response.error != null) {
            Response.error(ErrorType.valueOf(response.error!!.name), if (response.error == ErrorCode.BankReturnedError) null else response.errorMessage,
                if (response.error == ErrorCode.BankReturnedError && response.errorMessage !== null) listOf(response.errorMessage!!) else emptyList())
        } else {
            Response.error(ErrorType.UnknownError, response.errorMessage)
        }
    }

    private fun mapException(exception: Exception?): String? =
        exception?.stackTraceToString()

}